/* Copyright (C) 2005, 2006 Ted Phelps

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_LIBC_H
#include <libc.h>
#endif
#ifdef STDC_HEADERS
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif

#include "getopt.h"

/* We shouldn't rely on NAME_MAX being defined since it can vary
 * across filesystems.  On the other hand, if we allow this to vary
 * then things get very ugly very quickly, so I'm just going to rely
 * on people not using silly filenames */
#ifndef NAME_MAX
#define NAME_MAX 256
#endif

/* The size of the buffer to use for generating actual pathnames */
#define BUFFER_SIZE 256

/* The maximum number of symbolic links to follow when constructing a
 * canonical path */
#ifndef MAXSYMLINKS
#define MAXSYMLINKS 20
#endif

/* NeXTstep 3.3 doesn't define S_ISLNK, but does have _S_IFLNK */
#ifndef S_ISDIR
# define S_ISDIR(mode) (((mode) & (S_IFMT)) == (S_IFDIR))
#endif
#ifndef S_ISLNK
# define S_ISLNK(mode) (((mode) & (S_IFMT)) == (S_IFLNK))
#endif

/* Macro for testing if two components are equal */
#define COMP_EQ(c1, c2) \
  ((c1)->length == (c2)->length && \
   memcmp((c1)->component, (c2)->component, (c1)->length) == 0)


/* The path_t data type represents a Unix-style pathname as an array
 * of strings, making certain kinds of manipulations simpler */
typedef struct path path_t;
struct path
{
    /* This component's parent */
    path_t *parent;

    /* The length of this component */
    int length;

    /* This component of the path */
    char component[NAME_MAX];
};


/* Forward declarations */
static int
unstow_contents(path_t *target,
                path_t *updirs,
                path_t *relative_path,
                path_t *remainder,
                path_t *package_out);


static struct option long_options[] =
{
    { "no", no_argument, NULL, 'n' },
    { "conflicts", no_argument, NULL, 'c' },
    { "dir", required_argument, NULL, 'd' },
    { "target", required_argument, NULL, 't' },
    { "verbose", optional_argument, NULL, 'v' },
    { "delete", no_argument, NULL, 'D' },
    { "restow", no_argument, NULL, 'R' },
    { "version", no_argument, NULL, 'V' },
    { "help", no_argument, NULL, 'h' },
    { NULL, 0, NULL, 0 }
};

/* The name of the executable; basename of argv[0] */
static char *progname;

/* Just look for conflicts (implies not_really) */
static int conflicts = 0;

/* Don't change anything */
static int not_really = 0;

/* Verbosity level */
static int verbosity = 0;

/* The root directory of the stow tree (/usr/local/stow) */
static path_t *stow_root = NULL;

/* The root directory of the target tree (/usr/local) */
static path_t *target_root = NULL;

/* The number of packages */
static int package_count;

/* The packages */
static path_t **packages;


/* Ancient *BSD systems lack getwd */
#ifndef HAVE_GETCWD
static char *getcwd(char *buf, size_t size)
{
    char buffer[MAXPATHLEN];

    if (getwd(buffer) == NULL) {
        return NULL;
    }

    memcpy(buf, buffer, size);
    buf[size] = '\0';
    return buf;
}
#endif

/* Print if the verbosity level is high enough */
static void
xlog(int level, const char *format, ...)
{
    va_list ap;

    if (verbosity >= level) {
        va_start(ap, format);
        vfprintf(stderr, format, ap);
        va_end(ap);
    }
}

static void
perr(const char *format, ...)
{
    va_list ap;

    fprintf(stderr, "%s: error: ", progname);
    va_start(ap, format);
    vfprintf(stderr, format, ap);
    va_end(ap);
    fprintf(stderr, ": %s\n", strerror(errno));
}

/* Malloc or die trying */
static void *
xmalloc(size_t size)
{
    void *result;

    if ((result = malloc(size)) == NULL) {
	perr("unable to allocate %d bytes of memory", size);
	exit(1);
    }

    return result;
}

/* Realloc or die trying */
static void *
xrealloc(void *ptr, size_t size)
{
    void *result;

    if ((result = realloc(ptr, size)) == NULL) {
	perr("unable to allocate %d bytes of memory", size);
	exit(1);
    }

    return result;
}

/* Calloc or die trying */
static void *
xcalloc(size_t nmemb, size_t size)
{
    void *result;

    if ((result = calloc(nmemb, size)) == NULL) {
	perr("unable to allocate %d bytes of memory", nmemb * size);
	exit(1);
    }

    return result;
}

/* Initializes a chunk of memory as a path_t */
static void
path_init(path_t *self, path_t *parent, const char *component, int length)
{
    /* Sanity check */
    assert(length <= NAME_MAX);

    /* Fill in the fields */
    self->parent = parent;
    self->length = length;
    memcpy(self->component, component, length);
}

/* Allocates a path with room for capacity components */
static path_t *
path_alloc(path_t *parent, const char *component, int length)
{
    path_t *self;

    /* Allocate just enough room for the component */
    self = xmalloc(sizeof(path_t) - NAME_MAX + length);

    /* Initialize the fields */
    path_init(self, parent, component, length);
    return self;
}

/* Allocates a copy of the first count components of path.  If count
 * is negative then the entire path is duplicated */
static path_t *
path_clone(path_t *self, path_t *parent, int count)
{
    path_t *path;

    /* Watch for the end of the recursion */
    if (count == 0 || self == NULL) {
	return parent;
    }

    /* Clone the parent */
    path = path_clone(self->parent, parent, count < 0 ? count : count - 1);

    /* Clone and return this component */
    return path_alloc(path, self->component, self->length);
}

/* Frees a path and its parents */
static void
path_free(path_t *self)
{
    path_t *parent;

    while (self != NULL) {
	parent = self->parent;
	free(self);
	self = parent;
    }
}

/* Allocates a path_t from a string */
static path_t *
path_from_string(const char *pathname)
{
    path_t *self = NULL;
    path_t *parent;
    const char *point;
    const char *mark;

    /* Is this an absolute path? */
    point = pathname;
    if (*point == '/') {
	/* Record it as such with an empty first component */
	self = path_alloc(NULL, "/", 0);
    } else {
	self = NULL;
    }

    /* Split up the rest of the path */
    for (;;) {
	/* Skip any additional slashes */
	while (*point == '/') {
	    point++;
	}

	/* Is this the end of the path */
	if (*point == '\0') {
	    return self;
	}

	/* Find the end of the component */
	mark = point;
	while (*point != '/' && *point != '\0') {
	    point++;
	}

	/* Allocate new component */
	parent = self;
	if ((self = path_alloc(parent, mark, point - mark)) == NULL) {
	    path_free(parent);
	    return NULL;
	}
    }
}
/* Returns the number of components in the path */
static int
path_count(path_t *self)
{
    int count = 0;

    while (self != NULL) {
	self = self->parent;
	count++;
    }

    return count;
}

/* Returns nonzero if self and path are identical */
static int
path_equals(path_t *self, path_t *path)
{
    while (path != NULL) {
	/* Return false if the paths are different lengths */
	if (self == NULL) {
	    return 0;
	}

	/* Return false if these components are not the same */
	if (!COMP_EQ(self, path)) {
	    return 0;
	}

	/* Try the parents */
	self = self->parent;
	path = path->parent;
    }

    /* Return false if the paths are different lengths */
    if (self != NULL) {
	return 0;
    }

    /* They're equal! */
    return 1;
}

/* Returns nonzero if self is the same as the concatenation of updirs,
 * relative_path, package and remainder */
static int
path_equals4(path_t *self,
             path_t *package,
             path_t *updirs,
             path_t *relative_path,
             path_t *remainder)
{
    /* First compare against the remainder */
    while (remainder != NULL) {
	/* Return false if the paths are different lengths */
	if (self == NULL) {
	    return 0;
	}

	/* Return false if these components are not the same */
	if (!COMP_EQ(self, remainder)) {
	    return 0;
	}

	/* Try the parents */
	self = self->parent;
	remainder = remainder->parent;
    }

    /* Next we compare the package */
    assert(package->parent == NULL);

    /* Return false if there's no package */
    if (self == NULL) {
	return 0;
    }

    /* Return false if the package doesn't match */
    if (!COMP_EQ(self, package)) {
	return 0;
    }

    self = self->parent;

    /* Next try the relative_path */
    while (relative_path != NULL) {
	/* Return false if the paths are different lengths */
	if (self == NULL) {
	    return 0;
	}

	/* Return false if the components are not the same */
	if (!COMP_EQ(self, relative_path)) {
	    return 0;
	}

	/* Try the parents */
	self = self->parent;
	relative_path = relative_path->parent;
    }

    /* Then try the updirs */
    while (updirs != NULL) {
	/* Return false if the paths are different lengths */
	if (self == NULL) {
	    return 0;
	}

	/* Return false if the components are not the same */
	if (!COMP_EQ(self, updirs)) {
	    return 0;
	}

	/* Try the parents */
	self = self->parent;
	updirs = updirs->parent;
    }

    /* Return false if the paths are different lengths */
    if (self != NULL) {
	return 0;
    }

    /* It's a match! */
    return 1;
}

/* Helper function for path_to_string() */
static char *
path_to_string2(path_t *self, char **point, int length, char *buffer, int size)
{
    char *result;
    int use_slash;

    /* If this is the end of the path then allocate the memory */
    if (self == NULL) {
	/* If the buffer is big enough then just use it */
	if ((length + 1) < size) {
	    result = buffer;
	} else {
	    result = xmalloc(length + 1);
	}

	*point = result;
	return result;
    }

    /* Are we going to need a slash? */
    use_slash = length != 0 || self->length == 0;

    /* Delegate to our parent to allocate some memory */
    result = path_to_string2(self->parent,
                             point,
                             length + self->length + use_slash,
                             buffer, size);

    /* Insert our component */
    memcpy(*point, self->component, self->length);
    *point += self->length;

    /* Append a slash if we're not the last component */
    if (use_slash) {
	**point = '/'; (*point)++;
    }


    return result;
}

/* Allocates and returns a string representation of the path */
static char *
path_to_string(path_t *self, char *buffer, int size)
{
    char *point;
    char *result;

    /* Delegate */
    if ((result = path_to_string2(self, &point, 0, buffer, size)) != NULL) {
	*point = '\0';
    }

    return result;
}

/* Allocates a path_t *representing the current directory */
static path_t *
path_from_cwd(void)
{
    char *buffer = NULL;
    size_t size = BUFFER_SIZE;
    path_t *self;

    /* Get the working directory */
    for (;;) {
	buffer = xrealloc(buffer, size);

	/* Try to read the current working directory */
	if (getcwd(buffer, size) != NULL) {
	    self = path_from_string(buffer);
	    free(buffer);
	    return self;
	}

	/* Bail if something goes wrong */
	if (errno != ERANGE) {
	    free(buffer);
	    return NULL;
	}

	/* Try again with a larger buffer */
	size *= 2;
    }
}

/* Allocates a path_t *based on the contents of the symbolic link */
static path_t *
path_from_symlink(path_t *self)
{
    char path_buffer[BUFFER_SIZE];
    char *string;
    char *buffer = NULL;
    size_t size = BUFFER_SIZE;
    int length;

    /* Get our pathname */
    string = path_to_string(self, path_buffer, BUFFER_SIZE);

    /* Read the whole symlink */
    for (;;) {
	buffer = xrealloc(buffer, size);

	/* Try to read the link */
	if ((length = readlink(string, buffer, size - 1)) < 0) {
	    /* If the name was too long then try again with a bigger buffer */
	    if (errno == ENAMETOOLONG) {
		size *= 2;
		continue;
	    }

	    /* Otherwise we crash and burn */
	    perr("cannot read link %s", string);
	    exit(1);
	}

	/* Null-terminate the symlink name */
	buffer[length] = '\0';

	/* Construct a path from it */
	self = path_from_string(buffer);

	/* Clean up */
	if (path_buffer != string) {
	    free(string);
	}

	free(buffer);
	return self;
    }
}

/* Returns a pointer to the package component of the relative path
 * located in the given directory */
static path_t *
path_package(path_t *self, path_t *dir, path_t *stow_root,
             path_t **updirs_out, path_t **remainder_root_out)
{
    path_t *component, *package, *child;
    int count_dir = path_count(dir);
    int count_stow = path_count(stow_root);
    int count_self;
    int count_updirs;
    int count, i;

    /* Figure out how many '..' components are at the beginning of the path */
    count_updirs = 0;
    count_self = 0;
    child = self;
    for (component = self; component != NULL; component = component->parent) {
	count_self++;

	/* If this is an absolute path then it has no package */
	if (component->length == 0) {
	    return NULL;
	}

	/* Is this a '..' component? */
	if (component->length == 2 &&
	    component->component[0] == '.' &&
	    component->component[1] == '.') {
	    /* Yes.  Increase the count. */
	    count_updirs++;
	} else {
	    /* No.  Reset the count */
	    child = component;
	    count_updirs = 0;
	}
    }

    /* Record the updirs */
    if (updirs_out) {
	*updirs_out = child->parent;
    }

    /* Figure out which component should be the package */
    count = count_self - count_updirs + count_dir - count_updirs - count_stow;

    /* Bail if the link can't possibly point into the stow tree */
    if (count_self - count_updirs < count) {
	return NULL;
    }

    /* Find the package component */
    package = self;
    child = NULL;
    for (i = 0; i < count - 1; i++) {
	child = package;
	package = package->parent;
    }

    /* Record the child of the package */
    if (remainder_root_out) {
	*remainder_root_out = child;
    }

    /* Make sure the rest of the path matches the stow directory */
    component = package->parent;
    for (i = 0; i < count_self - count_updirs - count; i++) {
	if (component == NULL || !COMP_EQ(component, stow_root)) {
	    return NULL;
	}

	/* Move on to the parent nodes */
	if (stow_root->length != 0) {
	    stow_root = stow_root->parent;
	}

	component = component->parent;
    }

    /* Skip the elided bits of dir */
    for (i = 0; i < count_updirs; i++) {
	if (dir->length == 0) {
	    break;
	}

	dir = dir->parent;
    }

    /* Compare the rest of dir with the rest of stow */
    if (!path_equals(dir, stow_root)) {
	return NULL;
    }

    return package;
}

/* Allocates and returns the path from self to path */
static path_t *
path_relative_path_to(path_t *self, path_t *path)
{
    int count_self = path_count(self);
    path_t *path1 = self;
    int count1 = count_self;
    int count_path = path_count(path);
    path_t *path2 = path;
    int count2 = count_path;
    int count_parent;
    path_t *updirs;
    int i;

    /* Make sure path1 is no longer than path2 */
    while (count2 < count1) {
	path1 = path1->parent;
	count1--;
    }

    /* Make sure path2 is no longer than path1 */
    while (count1 < count2) {
	path2 = path2->parent;
	count2--;
    }

    /* The paths are now the same length.  Check where they start diverging */
    count_parent = count1;
    while (path1 != NULL) {
	if (!COMP_EQ(path1, path2)) {
	    count_parent = count1 - 1;
	}

	count1--;
	path1 = path1->parent;
	path2 = path2->parent;
    }

    /* Allocate the updirs */
    updirs = NULL;
    for (i = 0; i < count_self - count_parent; i++) {
	updirs = path_alloc(updirs, "..", 2);
    }

    /* Clone the remainder */
    return path_clone(path, updirs, count_path - count_parent);
}

/* Stat the file pointed to by self */
static int
path_lstat(path_t *self, struct stat *statbuf)
{
    char buffer[BUFFER_SIZE];
    char *string;
    int result;

    /* Convert the path into a string */
    string = path_to_string(self, buffer, BUFFER_SIZE);

    /* Stat the file */
    result = lstat(string, statbuf);

    /* Clean up */
    if (string != buffer) {
	free(string);
    }

    return result;
}

/* Stat the file pointed to by self */
static int
path_stat(path_t *self, struct stat *statbuf)
{
    char buffer[BUFFER_SIZE];
    char *string;
    int result;

    /* Convert the path into a string */
    string = path_to_string(self, buffer, BUFFER_SIZE);

    /* Stat the file */
    result = stat(string, statbuf);

    /* Clean up */
    if (string != buffer) {
	free(string);
    }

    return result;
}

/* Allocates an absolute path to the same file as path which does not
 * contain ., .. or any symbolic links.  The count indicates how many
 * symbolic links have been encountered thus far.  The path should be
 * grafted onto the beginning of the path is relative.  If path is
 * NULL then the current directory is used. */
static path_t *
path_canonicalize2(path_t *self, int *count, path_t *path)
{
    path_t *parent, *result;
    struct stat statbuf;
    path_t *link;

    /* Is the path unrooted? */
    if (self == NULL) {
	/* If there's a parent path then use it */
	if (path != NULL) {
	    return path;
	}

	/* Otherwise we'll use the current directory */
	self = path_from_cwd();
    }

    /* The root component ends the recursion */
    if (self->length == 0) {
	/* Sanity check */
	assert(self->parent == NULL);

	/* If there's a parent path then free it */
	if (path != NULL) {
	    path_free(path);
	}

	/* And return a new root node */
	return path_alloc(NULL, "/", 0);
    }

    /* Get a canonical version of the path up to here */
    if ((parent = path_canonicalize2(self->parent, count, path)) == NULL) {
	return NULL;
    }

    /* Is this the a '.' component? */
    if (self->length == 1 && self->component[0] == '.') {
	/* Yes.  Elide it. */
	return parent;
    }

    /* Is this a '..' component? */
    if (self->length == 2 && self->component[0] == '.' && self->component[1] == '.') {
	/* Is our parent node the root? */
	if (parent->length == 0) {
	    /* Yes.  Just elide this component. */
	    return parent;
	}

	/* Otherwise elide this component and our parent */
	result = parent->parent;
	free(parent);
	return result;
    }

    /* Duplicate this component */
    result = path_alloc(parent, self->component, self->length);

    /* Attach the parent node */
    result->parent = parent;

    /* Stat the path up to and including this component */
    if (path_lstat(result, &statbuf) < 0) {
	path_free(result);
	return NULL;
    }

    /* Is this a symbolic link? */
    if (S_ISLNK(statbuf.st_mode)) {
	/* Limit the number of symbolic links we'll follow */
	if (!((*count)++ < MAXSYMLINKS)) {
	    path_free(result);
	    errno = ELOOP;
	    return NULL;
	}

	/* Look up the symlink */
	link = path_from_symlink(result);

	/* Clean up */
	free(result);

	/* Canonicalize they symlink relative to our parent directory */
	result = path_canonicalize2(link, count, parent);

	/* Clean up */
	path_free(link);
    }

    /* Otherwise we're canonical as is */
    return result;
}

/* Front-end function for path_canonicalize2 */
static path_t *
path_canonicalize(path_t *self)
{
    int count = 0;
    return path_canonicalize2(self, &count, NULL);
}

/* Allocates a path_t *based on the real path denoted by string */
static path_t *
path_from_string_canonical(const char *pathname)
{
    path_t *path, *result;

    /* Create a path from the pathname */
    path = path_from_string(pathname);

    /* Get a canonical version of it */
    result = path_canonicalize(path);

    /* Clean up */
    path_free(path);
    return result;
}

/* Reads the contents of a directory and stores in in a dynamically
 * allocated array of paths */
static int
path_contents(path_t *self, path_t **files_out)
{
    char buffer[BUFFER_SIZE];
    char *string;
    DIR *dir;
    int count;
    struct dirent *dentry;

    /* Convert the path to a string */
    string = path_to_string(self, buffer, BUFFER_SIZE);

    /* Open the directory */
    if ((dir = opendir(string)) == NULL) {
	perr("cannot read directory %s", string);
	exit(1);
    }

    /* Read its contents */
    count = 0;
    *files_out = NULL;

    for (;;) {
	/* Read the next file */
	if ((dentry = readdir(dir)) == NULL) {
	    break;
	}

	/* Skip '.' and '..' */
	if (dentry->d_name[0] == '.' &&
	    (dentry->d_name[1] == '\0' ||
	     (dentry->d_name[1] == '.' && dentry->d_name[2] == '\0'))) {
	    continue;
	}

	/* Make more room in our table */
	*files_out = xrealloc(*files_out, (count + 1) * sizeof(path_t));

	/* Copy the path into place */
	path_init(*files_out + count++, self, dentry->d_name, NAMLEN(dentry));
    }

    /* Close the directory */
    if (closedir(dir) < 0) {
        perr("closedir failed");
	exit(1);
    }

    return count;
}

/* Creates a symbolic link pointing to path */
static void
path_symlink(path_t *self, path_t *path)
{
    char self_buffer[BUFFER_SIZE];
    char path_buffer[BUFFER_SIZE];
    char *self_string;
    char *path_string;

    /* Convert the paths to strings */
    self_string = path_to_string(self, self_buffer, BUFFER_SIZE);
    path_string = path_to_string(path, path_buffer, BUFFER_SIZE);

    /* Make some noise */
    xlog(1, "LINK %s to %s\n", self_string, path_string);

    /* Make the link */
    if (!not_really) {
	if (symlink(path_string, self_string) < 0) {
	    perr("could not create symlink %s to %s", path_string, self_string);
	    exit(1);
	}
    }

    /* Clean up */
    if (self_string != self_buffer) {
	free(self_string);
    }

    if (path_string != path_buffer) {
	free(path_string);
    }
}

/* Creates a directory */
static void
path_mkdir(path_t *self)
{
    char buffer[BUFFER_SIZE];
    char *string;

    /* Convert the path into a string */
    string = path_to_string(self, buffer, BUFFER_SIZE);

    /* Make some noise */
    xlog(1, "MKDIR %s\n", string);

    /* Make the directory */
    if (!not_really) {
	if (mkdir(string, 0777) < 0) {
	    perr("could not create directory %s", string);
	    exit(1);
	}
    }

    /* Clean up */
    if (string != buffer) {
	free(string);
    }
}

/* Creates a symbolic link pointing to the concatenation of updirs,
 * relative_path, package and remainder */
static void
path_symlink4(path_t *self,
              path_t *package,
              path_t *updirs,
              path_t *relative_path,
              path_t *remainder)
{
    path_t *relative_path_first;
    path_t *remainder_first;

    /* Find the beginning of the relative_path */
    relative_path_first = relative_path;
    while (relative_path_first->parent != NULL) {
	relative_path_first = relative_path_first->parent;
    }

    /* The package should not have a parent node */
    assert(package->parent == NULL);

    /* Find the beginning of the remainder */
    remainder_first = remainder;
    while (remainder_first->parent != NULL) {
	remainder_first = remainder_first->parent;
    }

    /* Concatenate the updirs, relative-path, package and remainder */
    relative_path_first->parent = updirs;
    package->parent = relative_path;
    remainder_first->parent = package;

    /* Create a link which points to the concatenation */
    path_symlink(self, remainder);
    
    /* Put everything back the way it was */
    relative_path_first->parent = NULL;
    package->parent = NULL;
    remainder_first->parent = NULL;
}

/* Unlinks the file pointed to by self */
static void
path_unlink(path_t *self)
{
    char buffer[BUFFER_SIZE];
    char *string;

    /* Convert the path to a string */
    string = path_to_string(self, buffer, BUFFER_SIZE);

    /* Make some noise */
    xlog(1, "UNLINK %s\n", string);

    /* Unlink the file */
    if (!not_really) {
	if (unlink(string) < 0) {
	    perr("could not unlink %s", string);
	    exit(1);
	}
    }

    /* Clean up */
    if (string != buffer) {
	free(string);
    }
}

/* Remove a directory */
static void
path_rmdir(path_t *self)
{
    char buffer[BUFFER_SIZE];
    char *string;

    /* Convert the path to a string */
    string = path_to_string(self, buffer, BUFFER_SIZE);

    /* Make some noise */
    xlog(1, "RMDIR %s\n", string);

    /* Remove the directory */
    if (!not_really) {
	if (rmdir(string) < 0) {
	    perr("could not remove directory %s", string);
	    exit(1);
	}
    }

    /* Clean up */
    if (string != buffer) {
	free(string);
    }
}

/* Recursively empties a directory of all symbolic links and subdirectories */
static void
path_purge(path_t *self)
{
    int count, i;
    path_t *files;
    struct stat statbuf;
    char buffer[BUFFER_SIZE];
    char *string;

    /* Get the directory contents */
    count = path_contents(self, &files);

    /* Go through each file in the directory */
    for (i = 0; i < count; i++) {
	/* Stat the file */
	if (path_lstat(files + i, &statbuf) < 0) {
            string = path_to_string(files + i, buffer, BUFFER_SIZE);
            perr("unable to stat file %s", string);
            if (string != buffer) {
                free(string);
            }

	    exit(1);
	}

	/* Delete symbolic links outright */
	if (S_ISLNK(statbuf.st_mode)) {
	    path_unlink(files + i);
	    continue;
	}

	/* Recursively delete directories */
	if (S_ISDIR(statbuf.st_mode)) {
	    /* Empty the directory */
	    path_purge(files + i);

	    /* Remove it */
	    path_rmdir(files + i);
	    continue;
	}

	/* Choke and die on normal files */
	fprintf(stderr, "%s: there's trouble captain!\n", progname);
	exit(1);
    }

    /* Clean up */
    if (files != NULL) {
	free(files);
    }
}




/* Print out a usage message */
static void
usage(int argc, char *argv[])
{
    fprintf(stderr,
            "%s (New Stow) version " VERSION "\n\n"
            "Usage: %s [OPTION]... PACKAGE...\n"
            "  -n, --no              Do not actually make changes\n"
	    "  -c, --conflicts       Scan for conflicts, implies -n\n"
	    "  -d DIR, --dir=DIR     Set stow dir to DIR (default is current dir)\n"
	    "  -t DIR, --target=DIR  Set target to DIR (default is parent of stow dir)\n"
	    "  -v, --verbose[=N]     Increase verboseness (levels are 0,1,2,3;\n"
	    "                          -v or --verbose adds 1; --verbose=N sets level)\n"
	    "  -D, --delete          Unstow instead of stow\n"
	    "  -R, --restow          Restow (like stow -D followed by stow)\n"
	    "  -V, --version         Show Stow version number\n"
	    "  -h, --help            Show this help\n",
            progname, progname);
}

/* Print out a message about a conflict */
static void
conflict(path_t *path1, path_t *path2)
{
    char buffer1[BUFFER_SIZE];
    char buffer2[BUFFER_SIZE];
    char *string1;
    char *string2;

    /* Convert the paths into strings */
    string1 = path_to_string(path1, buffer1, BUFFER_SIZE);
    string2 = path_to_string(path2, buffer2, BUFFER_SIZE);

    /* Print a helpful message */
    if (conflicts) {
	xlog(0, "CONFLICT: %s vs. %s\n", string1, string2);
    } else {
	xlog(0, "%s: CONFLICT: %s vs. %s\n", progname, string1, string2);
	exit(1);
    }

    /* Clean up */
    if (string1 != buffer1) {
	free(string1);
    }

    if (string2 != buffer2) {
	free(string2);
    }
}

/* Stow the contents of a directory */
static void
stow_contents(path_t *stow,
              path_t *target,
              path_t *package,
              path_t *updirs,
              path_t *relative_path,
              path_t *remainder);

/* Stow a directory */
static void stow_dir(path_t *stow,
                     path_t *target,
                     path_t *package,
                     path_t *updirs,
                     path_t *relative_path,
                     path_t *remainder)
{
    struct stat statbuf;
    path_t sub_updirs;
    char buffer1[BUFFER_SIZE];
    char buffer2[BUFFER_SIZE];
    char *string1;
    char *string2;
    path_t *link;
    path_t *link_package;
    path_t *link_updirs;
    path_t *link_package_parent;
    path_t *link_package_child;
    path_t *link_stow;

    /* Be verbose */
    if (verbosity >= 2) {
	/* Print a helpful message */
	string1 = path_to_string(package, buffer1, BUFFER_SIZE);
	string2 = path_to_string(remainder, buffer2, BUFFER_SIZE);
	xlog(2, "Stowing directory %s/%s\n", string1, string2);

	/* Clean up */
	if (string1 != buffer1) {
	    free(string1);
	}

	if (string2 != buffer2) {
	    free(string2);
	}
    }

    /* Check out the target file */
    if (path_lstat(target, &statbuf) < 0) {
	/* It doesn't exist so make a symbolic link */
	path_symlink4(target, package, updirs, relative_path, remainder);
	return;
    }

    /* Is it a directory? */
    if (S_ISDIR(statbuf.st_mode)) {
	/* Update the updirs */
	path_init(&sub_updirs, updirs, "..", 2);

	/* Recursively stow our contents */
	stow_contents(stow, target, package, &sub_updirs, relative_path, remainder);
	return;
    }

    /* Is it a symlink? */
    if (S_ISLNK(statbuf.st_mode)) {
	/* Dereference the link */
	link = path_from_symlink(target);

	/* Figure out which package it points to */
	link_package = path_package(link,
                                    target->parent,
                                    stow_root,
                                    &link_updirs,
                                    &link_package_child);

	/* If it doesn't point into the stow tree then we have a conflict */
	if (link_package == NULL) {
	    conflict(stow, target);
	    path_free(link);
	    return;
	}

	/* Is this the symlink we're trying to create? */
	if (path_equals4(link, package, updirs, relative_path, remainder)) {
	    /* Be verbose */
	    if (verbosity >= 3) {
		/* Print out some useful information */
		string1 = path_to_string(target, buffer1, BUFFER_SIZE);
		string2 = path_to_string(stow, buffer2, BUFFER_SIZE);
		xlog(3, "%s already points to %s\n", string1, string2);

		/* Clean up */
		if (string1 != buffer1) {
		    free(string1);
		}

		if (string2 != buffer2) {
		    free(string2);
		}
	    }

	    /* Clean up */
	    path_free(link);
	    return;
	}

	/* Remember the package's parent */
	link_package_parent = link_package->parent;

	/* Temporarily graft the package onto the stow and copy it */
	link_package->parent = stow_root;
	link_stow = path_clone(link, NULL, -1);
	link_package->parent = link_package_parent;

	/* Stat the file */
	if (path_stat(link_stow, &statbuf) < 0) {
	    /* The symlink points to nowhere */
	    /* Note: stow-1.3.2 deletes the link, but I think this is wrong -TAP */
	    conflict(stow, target);
	    path_free(link_stow);
	    path_free(link);
	    return;
	}

	/* If it's a directory then burst it */
	if (S_ISDIR(statbuf.st_mode)) {
	    /* Remove the symbolic link */
	    path_unlink(target);

	    /* Make a directory instead */
	    path_mkdir(target);

	    /* Stow the contents of the other package */
	    link_package->parent = NULL;
	    link_package_child->parent = NULL;
	    path_init(&sub_updirs, link_updirs, "..", 2);
	    stow_contents(link_stow, target, link_package, &sub_updirs, relative_path, link);

	    /* Put everything back the way it was */
	    link_package_child->parent = link_package;
	    link_package->parent = link_package_parent;

	    /* Tweak our updirs */
	    path_init(&sub_updirs, updirs, "..", 2);

	    /* And stow the new package */
	    stow_contents(stow, target, package, &sub_updirs, relative_path, remainder);

	    /* Clean up */
	    path_free(link_stow);
	    path_free(link);
	    return;
	}

	/* The existing symlink points to a non-directory */
	conflict(stow, target);

	/* Clean up */
	path_free(link_stow);
	path_free(link);
	return;
    }

    /* Otherwise there's a normal file where it shouldn't be */
    conflict(stow, target);
}

/* Stow a non-directory */
static void
stow_nondir(path_t *stow,
            path_t *target,
            path_t *package,
            path_t *updirs,
            path_t *relative_path,
            path_t *remainder)
{
    struct stat statbuf;
    path_t *link;
    char buffer1[BUFFER_SIZE];
    char buffer2[BUFFER_SIZE];
    char *string1;
    char *string2;

    /* Does the target file exist? */
    if (path_lstat(target, &statbuf) < 0) {
	/* No.  Just create the symlink and we're done. */
	path_symlink4(target, package, updirs, relative_path, remainder);
	return;
    }

    /* Is it a symbolic link? */
    if (S_ISLNK(statbuf.st_mode)) {
	/* Dereference the symlink */
	link = path_from_symlink(target);

	/* Is it the link we were going to make ourselves? */
	if (path_equals4(link, package, updirs, relative_path, remainder)) {
	    /* Be verbose */
	    if (verbosity >= 3) {
		/* Print out a helpful message */
		string1 = path_to_string(target, buffer1, BUFFER_SIZE);
		string2 = path_to_string(stow, buffer2, BUFFER_SIZE);
		xlog(3, "%s already points to %s\n", string1, string2);

		/* Clean up */
		if (string1 != buffer1) {
		    free(string1);
		}

		if (string2 != buffer2) {
		    free(string2);
		}
	    }

	    /* Clean up */
	    path_free(link);
	    return;
	}
    }

    /* Anything else is a conflict */
    conflict(stow, target);
}

/* Stow the contents of a directory */
static void
stow_contents(path_t *stow,
              path_t *target,
              path_t *package,
              path_t *updirs,
              path_t *relative_path,
              path_t *remainder)
{
    int count, i;
    path_t *files;
    char buffer1[BUFFER_SIZE];
    char buffer2[BUFFER_SIZE];
    char *string1;
    char *string2;
    struct stat statbuf;
    path_t sub_stow;
    path_t sub_target;
    path_t sub_remainder;

    /* Be verbose */
    if (verbosity >= 2) {
	/* Print a helpful message */
	string1 = path_to_string(package, buffer1, BUFFER_SIZE);
	string2 = path_to_string(remainder, buffer2, BUFFER_SIZE);

	if (*string2 != '\0') {
	    xlog(2, "Stowing contents of %s/%s\n", string1, string2);
	} else {
	    xlog(2, "Stowing contents of %s\n", string1);
	}

	/* Clean up */
	if (string1 != buffer1) {
	    free(string1);
	}

	if (string2 != buffer2) {
	    free(string2);
	}
    }

    /* Read the contents of the directory */
    count = path_contents(stow, &files);

    /* Go through each of the files */
    for (i = 0; i < count; i++) {
	/* Get its file type */
	if (path_stat(files + i, &statbuf) < 0) {
            string1 = path_to_string(files + i, buffer1, BUFFER_SIZE);
            if (errno == ENOENT) {
                xlog(0, "%s: warning: link points to nonexistent file: %s\n",
                        progname, string1);
                /* Treat it as a normal file */
                statbuf.st_mode = 0;
            } else {
                perr("unable to stat file %s", string1);
                if (string1 != buffer1) {
                    free(string1);
                }

                exit(1);
            }
	}

	/* Update the various paths for the file */
	path_init(&sub_stow, stow, (files + i)->component, (files + i)->length);
	path_init(&sub_target, target, (files + i)->component, (files + i)->length);
	path_init(&sub_remainder, remainder, (files + i)->component, (files + i)->length);

	/* Delegate */
	if (S_ISDIR(statbuf.st_mode)) {
	    stow_dir(&sub_stow,
                     &sub_target,
                     package,
                     updirs,
                     relative_path,
                     &sub_remainder);
	} else {
	    stow_nondir(&sub_stow,
                        &sub_target,
                        package,
                        updirs,
                        relative_path,
                        &sub_remainder);
	}
    }

    /* Clean up */
    if (files != NULL) {
	free(files);
    }
}


/* Coalesce a directory into a single symlink or nothing */
static void
coalesce_dir(path_t *target,
             path_t *package,
             path_t *updirs,
             path_t *relative_path,
             path_t *remainder)
{
    /* Empty the target directory */
    path_purge(target);

    /* Remove the directory */
    path_rmdir(target);

    /* If there's still a package which refers the directory then
     * replace it with a symlink. */
    if (package->length != 0) {
        path_symlink4(target, package, updirs, relative_path, remainder);
    }
}

/* Unstow a symbolic link */
static int
unstow_nondir(path_t *target,
              path_t *updirs,
              path_t *relative_path,
              path_t *remainder,
              path_t *package_out)
{
    path_t *link_target;
    path_t *package;
    int i;

    /* Dereference the link */
    link_target = path_from_symlink(target);
    
    /* Get the package associated with that link */
    if ((package = path_package(link_target,
                                target->parent,
                                stow_root,
                                NULL, NULL)) == NULL) {
	path_free(link_target);
	return 0;
    }

    /* Is this one of the packages we're removing? */
    for (i = 0; i < package_count; i++) {
	if (COMP_EQ(package, packages[i])) {
	    path_unlink(target);

	    /* No package to return */
	    package_out->length = 0;

	    /* Clean up */
	    path_free(link_target);
	    return 1;
	}
    }

    /* Copy the package into place */
    path_init(package_out, NULL, package->component, package->length);

    /* Clean up */
    path_free(link_target);

    /* Not deleting it */
    return 1;
}

/* Unstow a directory */
static int
unstow_dir(path_t *target,
           path_t *updirs,
           path_t *relative_path,
           path_t *remainder,
           path_t *package_out)
{
    path_t sub_updirs;
    path_t sub_remainder;

    /* Add another '..' to the updirs */
    path_init(&sub_updirs, updirs, "..", 2);

    /* Add another directory to the remainder */
    path_init(&sub_remainder, remainder, target->component, target->length);

    /* Unstow the directory's contents */
    return unstow_contents(target,
                           &sub_updirs,
                           relative_path,
                           &sub_remainder,
                           package_out);
}

/* Unstow packages */
static int
unstow_contents(path_t *target,
                path_t *updirs,
                path_t *relative_path,
                path_t *remainder,
                path_t *package_out)
{
    path_t package;
    path_t component;
    path_t sub_remainder;
    path_t *parent;
    struct stat statbuf;
    char buffer[BUFFER_SIZE];
    char *string;
    int pure = 1;
    int count, i;
    path_t *files;
    int pure_count = 0;
    path_t *pure_packages = NULL;

    /* Don't go into the stow directory itself */
    if (path_equals(target, stow_root)) {
	/* The stow directory is not pure */
	return 0;
    }

    /* Check for a '.stow' file in the target directory */
    path_init(&component, target, ".stow", 5);
    if (!(path_lstat(&component, &statbuf) < 0)) {
	return 0;
    }

    /* Be verbose */
    if (verbosity >= 2) {
	/* Print a helpful message */
	string = path_to_string(target, buffer, BUFFER_SIZE);
	xlog(2, "Unstowing in %s\n", string);

	/* Clean up */
	if (string != buffer) {
	    free(string);
	}
    }

    /* Clear the result package */
    path_init(package_out, NULL, NULL, 0);

    /* Read the contents of the directory */
    count = path_contents(target, &files);

    /* Go through each of the files */
    for (i = 0; i < count; i++) {
	/* Get its file type */
	if (path_lstat(files + i, &statbuf) < 0) {
            string = path_to_string(files + i, buffer, BUFFER_SIZE);
            perr("unable to stat file %s", string);
            if (string != buffer) {
                free(string);
            }

	    exit(1);
	}

	/* Is it a symbolic link? */
	if (S_ISLNK(statbuf.st_mode)) {
	    /* Unstow the symlink */
	    if (unstow_nondir(files + i, updirs, relative_path, remainder, &package)) {
		/* If we're still pure then check for a new package */
		if (pure && package.length != 0) {
		    /* If no existing package then record this one */
		    if (package_out->length == 0) {
			path_init(package_out,
				  NULL,
				  package.component,
				  package.length);
		    } else if (!COMP_EQ(package_out, &package)) {
                        /* The package is not the same; we're no longer pure */
			pure = 0;
		    }
		}
	    } else {
                pure = 0;
            }

	    /* This file isn't a pure subdirectory */
	    continue;
	}

	/* Is it a directory */
	if (S_ISDIR(statbuf.st_mode)) {
	    /* Unstow the directory */
	    if (unstow_dir(files + i, updirs, relative_path, remainder, &package)) {
		/* Record the package and path so that we can coalesce
		 * later if this directory turns out to be impure */
		pure_packages = xrealloc(pure_packages,
                                         sizeof(path_t) * (pure_count + 1));
		path_init(pure_packages + pure_count,
			  files + i,
			  package.component,
			  package.length);
		pure_count++;

		/* The subdirectory is pure.  Is there a package in it? */
		if (package.length != 0) {
		    /* Have we found a package in this directory yet? */
		    if (package_out->length == 0) {
			/* No.  Just record the other package. */
			path_init(package_out,
                                  NULL,
                                  package.component,
                                  package.length);
		    } else {
			/* Yes.  Are they the same packages? */
			if (!COMP_EQ(package_out, &package)) {
			    /* Nope.  We're no longer a pure directory. */
			    pure = 0;
			}
		    }
		}
	    } else {
		/* If our subdir isn't pure then we aren't */
		pure = 0;
	    }

	    continue;
	}

	/* Normal files are considered 'impure' */
	pure = 0;
    }

    /* If this directory isn't pure or if it's the top-level directory
     * then we need to coalesce its pure subtrees */
    if (pure_count != 0 && (!pure || !remainder)) {
	/* Coalesce the directories that need it */
	for (i = 0; i < pure_count; i++) {
	    parent = pure_packages[i].parent;

	    /* Make a path pointing from the root of the stow tree to
	     * the target of our symlink */
	    path_init(&sub_remainder, remainder, parent->component, parent->length);

	    /* Temporarily remove the package's parent */
	    pure_packages[i].parent = NULL;
		      
	    /* Coalesce the directory */
	    coalesce_dir(parent, &pure_packages[i], updirs, relative_path, &sub_remainder);

	    /* Give the package its parent back */
	    pure_packages[i].parent = parent;
	}
    }

    /* Clean up */
    if (pure_packages != NULL) {
	free(pure_packages);
    }

    if (files != NULL) {
	free(files);
    }

    return pure;
}


/* Parse the command line args and arrange to [un]stow */
int
main(int argc, char *argv[])
{
    int do_delete = 0;
    int do_restow = 0;
    int choice;
    int i;
    path_t *relative_path;
    path_t package;
    path_t *path;
    char buffer[BUFFER_SIZE];
    char *string;

    /* Determine the program name */
    if ((progname = strrchr(argv[0], '/')) == NULL) {
        progname = argv[0];
    } else {
        progname++;
    }

    /* Parse the command-line options */
    while ((choice = getopt_long(argc, argv, "ncd:t:vDRVh", long_options, NULL)) != -1) {
	switch (choice) {
	    case 'n':
                /* Don't really stow */
		not_really = 1;
		break;

	    case 'c':
                /* Check for conflicts */
		conflicts = 1;
		not_really = 1;
		break;

	    case 'd':
                /* Set the stow directory.  First discard any previous
                 * stow root */
		if (stow_root != NULL) {
		    path_free(stow_root);
		}

		/* Create a path from the argument */
		if ((stow_root = path_from_string_canonical(optarg)) == NULL) {
		    perr("invalid stow directory");
		    exit(1);
		}

		break;

	    case 't':
                /* Set the target directory.  First discard any
                 * previous target root */
		if (target_root != NULL) {
		    path_free(target_root);
		}

		/* Create a path from the argument */
		if ((target_root = path_from_string_canonical(optarg)) == NULL) {
		    perr("invalid target directory");
		    exit(1);
		}

		break;

	    case 'v':
                /* Increase the verbosity */
		if (optarg == NULL) {
		    verbosity++;
		} else {
		    verbosity = atoi(optarg);
		}

		break;

	    case 'D':
                /* Unstow */
		do_delete = 1;
		break;

	    case 'R':
                /* Restow */
		do_restow = 1;
		break;

	    case 'V':
                /* Print the version and exit */
		printf("%s (New Stow) version " VERSION "\n", progname);
		exit(0);

	    case 'h':
                /* Print out help and exit */
		usage(argc, argv);
		exit(0);

	    case '?':
                /* Unknown option.  Since getopt_long has already
                 * printed an error message, just print usage and
                 * exit. */
		usage(argc, argv);
		exit(1);

	    default:
                /* Trouble */
                fprintf(stderr, "%s: error: unhandled option: -%c\n",
                        progname, choice);
		abort();
	}
    }

    /* Make sure we have something to stow */
    package_count = argc - optind;
    if (package_count == 0) {
	fprintf(stderr, "%s: error: no packages named\n", progname);
	usage(argc, argv);
	exit(1);
    }

    /* Make sure the stow directory is set */
    if (stow_root == NULL) {
	/* Default the the current directory */
	path = path_from_cwd();

	/* Canonicalize it */
	stow_root = path_canonicalize(path);
	path_free(path);
    }

    /* Make sure the target directory is set */
    if (target_root == NULL) {
	/* If the stow root is / then use that */
	if (stow_root->parent == NULL) {
	    target_root = path_clone(stow_root, NULL, -1);
	} else {
            /* Use the stow root's parent */
	    target_root = path_clone(stow_root->parent, NULL, -1);
	}
    }

    /* Figure out how to get from the target root to the stow root */
    relative_path = path_relative_path_to(target_root, stow_root);

    /* Allocate a table to store the packages */
    packages = xcalloc(package_count, sizeof(path_t));

    /* Convert the packages to stow into path components */
    for (i = 0; i < package_count; i++) {
	/* Convert the package into a path */
	packages[i] = path_from_string(argv[optind + i]);

	/* Make sure the path has only one component */
	if (packages[i]->parent != NULL || packages[i]->length == 0) {
	    fprintf(stderr, "%s: error: slashes not permitted in package names\n",
                progname);
	    exit(1);
	}
    }


    /* Should we unstow? */
    if (do_delete || do_restow) {
	unstow_contents(target_root, NULL, relative_path, NULL, &package);
    }

    /* Should we now stow? */
    if (!do_delete || do_restow) {
	for (i = 0; i < package_count; i++) {
	    /* Make some noise */
	    if (verbosity >= 1) {
		/* Print a helpful message */
		string = path_to_string(packages[i], buffer, BUFFER_SIZE);
		xlog(1, "Stowing package %s...\n", string);

		/* Clean up */
		if (string != buffer) {
		    free(string);
		}
	    }

	    /* Graft the package into the stow directory */
	    path_init(&package, stow_root, packages[i]->component, packages[i]->length);

	    /* Stow it */
	    stow_contents(&package, target_root, packages[i], NULL, relative_path, NULL);
	}
    }

    /* Clean up */
    for (i = 0; i < package_count; i++) {
	free(packages[i]);
    }

    free(packages);
    path_free(stow_root);
    path_free(target_root);
    path_free(relative_path);
    exit(0);
}

